import sys

def eval_expr(*args):
    if len(args) == 0:
        return "Empty expression..."
    else:
        try:
            expr = ' '.join(args)
            return eval(expr)
        except SyntaxError:
            print("Illegal expression: " + expr)
            sys.exit(1)



import curses
import curses.textpad

# a simple vim like editor
VIEW_STR = "[VIEW MODE]"
CMD_STR = "[CMD MODE]"
INSERT_STR = "[INSERT MODE]"
UNKNOWN_STR = "[UNKNOW STATE!!!]"

MOVE_KEYS = [curses.KEY_UP, curses.KEY_DOWN, curses.KEY_LEFT, curses.KEY_RIGHT]

# class Textbox(curses.textpad.Textbox):
#     def do_command(self, ch):
#         if ch == 127:
#             super(curses.textpad.Textbox, self).do_command(4)
#         else:
#             super(curses.textpad.Textbox, self).do_command(ch)


# heler function to get the code of key
def keyLookup():
    stdscr = curses.initscr()
    ch = stdscr.getch()
    curses.endwin()
    print(ch)


class TxtEditor:
    def __init__(self, content):
        self.content_ = content
        ## 0 is view mode, 1 is cmd mode, 2 is insert mode
        self.state_ = 0
        self.cmd_buff_ = ""
        self.stay_editing_ = True
        # cursor pos of cmd window
        self.yxc_ = [1, 0]
        # cursor pos of txt window
        self.yxt_ = [0, 0]


    def _render_state(self, window):
        if self.state_ == 0:
            stt_str = VIEW_STR
        elif self.state_ == 1:
            stt_str = CMD_STR
        elif self.state_ == 2:
            stt_str = INSERT_STR
        else:
            stt_str = UNKNOWN_STR
        # yx = curses.getsyx()

        window.addstr(0, 0, stt_str + '\n' + self.cmd_buff_)
        # window.move(*yx)


    def _render_content(self, window):
        yx = curses.getsyx()

        window.addstr(0, 0, '\n'.join(self.content_))
        #if self.state_ != 1:
        #    window.move(*yx)


    def _take_and_process_view(self, window):
        c = window.getch()
        if c == ord('i'):
            self.state_ = 2
        elif c == ord(':'):
            self.state_ = 1
            self.cmd_buff_ = ":"
        elif c in MOVE_KEYS:
            self._move_cursor(c)


    def _take_and_process_cmd(self, window):
        c = window.getch()
        if c >= 97 and c <= 122:
            self.cmd_buff_ = self.cmd_buff_ + chr(c)
            self.yxc_ = [2, len(self.cmd_buff_)]
        elif c == 10: # ENTER
            if self.cmd_buff_ == ":q":
                self.stay_editing_ = False
                self.cmd_buff_ = ""
        elif c == 27:
            self.cmd_buff_ = ""
            self.state_ = 0
        #window.move(2, len(self.cmd_buff_))


    def _take_and_process_insert(self, window):
        c = window.getch()
        self.stay_editing_ = True
        pass


    def _reset_state(self):
        self.state_ = 0
        self.cmd_buff_ = ""
        self.stay_editing_ = True


    def _move_cursor(self, move):
        assert move in MOVE_KEYS, "UNKNOWN MOVE KEYS"
        #yx = curses.getsyx()
        if move == MOVE_KEYS[0]:
            #window.move(max(0, yx[0] - 1), yx[1])
            self.yxt_[0] = max(0, yx[0] - 1)
        elif move == MOVE_KEYS[1]:
            #window.move(min(len(self.content_) - 1, yx[0] + 1), yx[1])
            #window.move(1, 1)
            self.yxt_[0] = min(len(self.content_) - 1,  yx[0] + 1)
        elif move == MOVE_KEYS[2]:
            #window.move(yx[0], max(0, yx[1] - 1))
            self.yxt_[1] = max(0, yx[1] - 1)
        elif move == MOVE_KEYS[3]:
            #window.move(yx[0], min(len(self.content_[yx[0]]), yx[1] + 1))
            self.yxt_[1] = min(len(self.content_[yx[0]]), yx[1] + 1)


    def run(self):
        self._reset_state()
        cmdscr = curses.initscr()
        txtscr = cmdscr.subwin(2, 0)
        cmdscr.clearok(True)
        txtscr.clearok(True)
        self._render_state(cmdscr)
        self._render_content(txtscr)
        while self.stay_editing_:
            cmdscr.move(*self.yxc_)
            txtscr.move(*self.yxt_)
            if self.state_ == 0:
                cmdscr.refresh()
                self._take_and_process_view(txtscr)
            elif self.state_ == 1:
                txtscr.refresh()
                self._take_and_process_cmd(cmdscr)
            elif self.state_ == 2:
                cmdscr.refresh()
                self._take_and_process_insert(txtscr)
            self._render_state(cmdscr)
            self._render_content(txtscr)
        cmdscr.clear()
        txtscr.clear()
        curses.endwin()
        self._reset_state()
        return self.content_



from __future__ import print_function
# -*- coding: utf-8 -*-

import calendar
import hashlib
import json
import os
import re
import time
from ada.txt_editor import LineFixedTxtEditor, InteractivePrompt
from getpass import getpass
from Crypto.Cipher import Blowfish

'''
let ada remember either a file, a string and peek the remembered content
for peek, ada will print LINE_BUFFER_SIZE lines above N_MORE_LINES_TO_PRINT lines below
around the regex keyword (case insensitive);
if no keyword provided, ada will show all the content

layout of config:
{file_alias0: file_path0, file_alias1: file_path1, ...}
'''
GREEN_CODE = u"\033[0;32m"
RED_CODE = u"\033[0;31m"
RESET_CODE = u"\033[0;0m"

START_LINE = ">>>>>>>>>>>>>>>>>>>>>>>>>>"
END_LINE = "<<<<<<<<<<<<<<<<<<<<<<<<<<"

LINE_BUFFER_SIZE = 10
N_MORE_LINES_TO_PRINT = 10

# default TTL is 15 days
DEFAULT_TTL = 15
DAY_SECONDS = 3600 * 24

_, col = os.popen('stty size', 'r').read().split()
EDITOR_WIDTH = int(col) - 1

CONFIG_PATH = os.path.expanduser('~/.adaconfig/.config')
MEMORY_PATH = os.path.expanduser('~/.adaconfig/.memory')
PWD_BLOCK_SIZE = 8
PWD_HEADING = "pwdbook:"

def process_memory(*args):
    try:
        # check if the config dir exists
        if os.path.exists(CONFIG_PATH):
            # load config json
            fp = open(CONFIG_PATH, 'r')
            config = json.load(fp)
            fp.close()
        else:
            os.makedirs(os.path.expanduser('~/.adaconfig'))
            fp = open(MEMORY_PATH, 'w')
            fp.close()
            # initilize the config json
            config = {"_memory": MEMORY_PATH}

        if 'peek' in args:
            idx = args.index('peek')
            peek(config, *args[idx + 1:])
        elif 'remember' in args:
            idx = args.index('remember')
            if idx == len(args) - 1:
                ed1 = LineFixedTxtEditor(EDITOR_WIDTH, [''])
                ctt = ed1.run()
                txt = ctt[0][0]
                ed2 = InteractivePrompt(EDITOR_WIDTH, "ttl (days): ")
                ctt2 = ed2.run()
                ttl_str = ctt2[0][0].replace(' ', '')
                try:
                    if len(ttl_str) > 0:
                        ttl = int(ttl_str)
                    else:
                        ttl = DEFAULT_TTL
                except Exception:
                    print("unrecognized ttl string {}, using default ttl {}".format(ttl_str, DEFAULT_TTL))
                    ttl = DEFAULT_TTL
                remember(config, txt, ttl)
            elif idx == len(args) - 2:
                remember(config, args[idx + 1])
            else:
                remember(config, args[idx + 1], int(args[-1]))
        elif 'register' in args:
            idx = args.index('register')
            if idx + 2 > len(args) - 1:
                print('to register, specify <alias> <file_path>')
                return
            register(config, args[idx + 1], args[idx + 2])
        elif 'forget' in args:
            idx = args.index('forget')
            if idx >= len(args) - 1:
                print('forget what?')
                return
            if args[idx + 1] in config:
                config.pop(args[idx + 1], None)
            else:
                # try deleting hash in _memory
                for hash in args[idx + 1:]:
                    forget(config, hash)
        elif 'config' in args:
            print(config)
        elif 'amend' in args:
            idx = args.index('amend')
            if len(args) - 1 < idx + 1:
                print('to amend, specify <hash> [<new_content>]')
            elif len(args) - 1 == idx + 1:
                hash = args[idx + 1]
                ret = forget(config, hash)
                if ret[1] <= 0:
                    print("hash not found, update nothing")
                else:
                    editor = LineFixedTxtEditor(EDITOR_WIDTH, [ret[2]])
                    ctt = editor.run()
                    editor = InteractivePrompt(EDITOR_WIDTH, "new ttl (days): ")
                    ctt2 = editor.run()
                    ttl_str = ctt2[0][0].replace(' ', '')
                    try:
                        if len(ttl_str) > 0:
                            ttl = int(ttl_str)
                        else:
                            ttl = ret[1]
                    except Exception:
                        print("unrecognized ttl string {}, using original ttl {}".format(ttl_str, ret[1]))
                        ttl = ret[1]
                    remember(config, '\n'.join(ctt[0]), ttl, ret[0])
            else:
                hash = args[idx + 1]
                new_content = ' '.join(args[idx + 2:])
                ret = forget(config, hash)
                if ret[1] <= 0:
                    print("hash not found, create new entry")
                    remember(config, new_content)
                else:
                    remember(config, new_content, ret[1], ret[0])
        elif 'append' in args:
            idx = args.index('append')
            if len(args) - 1 < idx + 2:
                print('to amend, specify <hash> <content_to_append>')
            else:
                hash = args[idx + 1]
                new_content = ' '.join(args[idx + 2:])
                ret = forget(config, hash)
                if ret[1] <= 0:
                    print("hash not found, append to nothing")
                else:
                    remember(config, ret[-1] + new_content, ret[1], ret[0])
        elif 'pwd' in args:
            idx = args.index('pwd')
            if len(args[idx+1:]) == 0:
                print('no action under pwd')
            else:
                pwd(config, *args[idx+1:])
        else:
            print("I don't understand the input" + ' '.join(args))
            return
        # flush the config, return
        fp = open(CONFIG_PATH, 'w')
        json.dump(config, fp)
    except Exception as e:
        print(e)


'''
args can be:
1. <empty>: showing all memory
2. [<alias/file/dir>]: showing all content in alias/file/dir
3. [<alias/file/dir> <regex>]: showing the adjacent lines around regex matches
4. [<regex>]: showing all memory that matches regex
'''
def peek(config, *args):
    # case 1
    if len(args) == 0:
        if os.path.exists(config['_memory']):
            with open(config['_memory'], 'r') as fp:
                entries = json.load(fp)
            MEMFORMAT = (GREEN_CODE + "[{} {}] " + RESET_CODE + "{} " + RED_CODE + "(expires in {} hrs)" + RESET_CODE)
            for entry in entries:
                ts = time.localtime(entry[0])
                time_str = "{}-{}-{} {}:{}:{}".format(
                    ts.tm_year,
                    str(ts.tm_mon).rjust(2, '0'),
                    str(ts.tm_mday).rjust(2, '0'),
                    str(ts.tm_hour).rjust(2, '0'),
                    str(ts.tm_min).rjust(2, '0'),
                    str(ts.tm_sec).rjust(2, '0'))
                now = calendar.timegm(time.gmtime())
                if entry[3] * DAY_SECONDS - (now - entry[0]) > 0:
                    print(MEMFORMAT.format(time_str, entry[2], transformBulletsView(entry[1]), "{:.2f}".format((entry[3] * DAY_SECONDS - (now - entry[0])) / 3600.0)))
    # case 2 or 3
    # TODO(yuquanshan): refactor code to two functions: _process_file and _process_dir
    elif args[0] in config or os.path.exists(os.path.expanduser(args[0])):
        # case 2
        good = False
        if len(args) == 1:
            if args[0] in config:
                with open(config[args[0]], 'r') as fp:
                    print(fp.read(), end="")
                    good = True
            else:
                path = os.path.expanduser(args[0])
                if os.path.exists(path):
                    if os.path.isfile(path):
                        with open(path, 'r') as fp:
                           print(fp.read(), end="")
                        good = True
                    elif os.path.isdir(path):
                        file_paths = [os.path.join(path, p) for p in os.listdir(path) if os.path.isfile(os.path.join(path, p))]
                        for p in file_paths:
                            print(RED_CODE + "<{}>".format(p)+ RESET_CODE)
                            with open(p, 'r') as fp:
                                print(fp.read(), end="")
                        good = True
            if not good:
                print("ERROR:" + str(args[0]) + "is neither a registered alias nor a valid path to a file or dir")
        # case 3
        else:
            pattern = u".*{}.*".format(args[1]).lower()
            print(RED_CODE + u"SEARCH PATTERN (case insensitive): " + pattern + RESET_CODE)
            if args[0] in config:
                with open(config[args[0]], 'r') as fp:
                    lines = fp.readlines()
                    print_match(lines, pattern)
                    good = True
            else:
                path = os.path.expanduser(args[0])
                if os.path.exists(path):
                    if os.path.isfile(path):
                        with open(path, 'r') as fp:
                            lines = fp.readlines()
                            print_match(lines, pattern)
                            good = True
                    elif os.path.isdir(path):
                        file_paths = [os.path.join(path, p) for p in os.listdir(path) if os.path.isfile(os.path.join(path, p))]
                        for p in file_paths:
                            print(RED_CODE + "<{}>".format(p)+ RESET_CODE)
                            with open(p, 'r') as fp:
                                lines = fp.readlines()
                                print_match(lines, pattern)
                                good = True
            if not good:
                print("ERROR:" + str(args[0]) + "is neither a registered alias nor a valid path to a file or dir")
    # case 4
    else:
        with open(config['_memory'], 'r') as fp:
            pattern = u".*{}.*".format(args[0]).lower()
            print(RED_CODE + u"SEARCH PATTERN (case insensitive): " + pattern + RESET_CODE)
            entries = json.load(fp)
        lines = []
        for entry in entries:
            lines.append(entry[1] + '\n')
        print_match(lines, pattern)


def remember(config, s, ttl = DEFAULT_TTL, ts = None):
    entries = []
    if os.path.exists(config['_memory']):
        with open(config['_memory'], 'r') as f:
            if len(f.read()) > 0:
                f.seek(0)
                entries = json.load(f)
    if not ts:
        ts = calendar.timegm(time.gmtime())
    buffered_entries = []
    hash = hashlib.sha224(s.encode()).hexdigest()[:6]
    hash_set = set()
    for entry in entries:
        tdelta = ts - entry[0]
        if tdelta > entry[3] * DAY_SECONDS:
            # ttl passed, forget this entry
            continue
        else:
            hash_set.add(entry[2])
            if hash in hash_set:
                print("WARNING: hash collides with an exising content, replace it with the new one")
                continue
            buffered_entries.append(entry)
    buffered_entries.append([ts, s, hash, ttl])
    with open(config['_memory'], 'w') as f:
        json.dump(buffered_entries, f)


# forget a _memory entry in hash, return [ts, ttl, original_content] ,
# if ttl <= 0, then either the entry is expired or non-existing
def forget(config, hash):
    entries = []
    ret = [0, 0, '']
    if os.path.exists(config['_memory']):
        with open(config['_memory'], 'r') as f:
            if len(f.read()) > 0:
                f.seek(0)
                entries = json.load(f)
    buffered_entries = []
    for entry in entries:
        ety = entry
        if ety[2] == hash:
            # if hash matches, forget by not adding to buffered_entries
            ret = [ety[0], ety[-1], ety[1]]
            continue
        else:
            buffered_entries.append(ety)
    with open(config['_memory'], 'w') as f:
        json.dump(buffered_entries, f)
    return ret


def register(config, alias, fp):
    if not os.path.exists(os.path.expanduser(fp)):
        raise Exception("File " + fp + " doesn't exist")
    config[alias] = fp


def pwd(config, action, *args):
    pwd = getpass('Enter you password: ')
    crypt = Blowfish.new(pwd)
    # schema: pwdbook:{name: [uname, pwd, comment]...}
    pwd_list = {}
    if '_pwd' not in config:
        print("please register _pwd first")
    if os.path.exists(config['_pwd']):
        with open(config['_pwd'], 'rb') as f:
            txt = f.read()
            if len(txt) >= PWD_BLOCK_SIZE:
                txt = crypt.decrypt(txt).decode('ASCII')
                if PWD_HEADING not in txt:
                    raise Exception('Wrong password: ' + pwd)
                txt = txt.replace(PWD_HEADING, '', 1).strip()
                pwd_list = json.loads(txt)
    if action == 'search':
        if len(args) == 0:
            print('nothing to peek')
            return
        kw = args[0]
        found = False
        for i in pwd_list.items():
            if kw in i[0].lower():
                found = True
                print("{}: {}".format(i[0], i[1]))
        if not found:
            print(kw + ' not found')
    elif action == 'update':
        if len(args) < 3:
            print('please at least specify name, uname, and pwd')
            return
        name = args[0]
        ctnt = args[1:]
        pwd_list[name] = ctnt
    elif action == 'delete':
        if len(args) == 0:
            print('nothing to delete')
            return
        if args[0] not in pwd_list:
            print('cannot find ' + args[0])
            return
        del pwd_list[args[0]]
    txt = json.dumps(pwd_list)
    txt = str(PWD_HEADING)+txt
    txt = txt.ljust((len(txt) // PWD_BLOCK_SIZE  + 1) * PWD_BLOCK_SIZE)
    etxt = crypt.encrypt(txt)
    with open(config['_pwd'], 'wb') as f:
        f.write(etxt)


def print_match(lines, pattern):
    line_buffer = []
    match_in_buffer = False
    more_lines_countdown = 0
    line_count = 0
    for l in lines:
        line_count = line_count + 1
        l = l
        match = re.match(pattern, l.lower())
        if match:
            match_in_buffer = True
            more_lines_countdown = N_MORE_LINES_TO_PRINT
            line_buffer.append(GREEN_CODE + l + RESET_CODE)
        elif match_in_buffer:
            line_buffer.append(l)
            more_lines_countdown = more_lines_countdown - 1
            if more_lines_countdown == 0:
                print(START_LINE)
                for i in line_buffer:
                    print(i, end="")
                print(END_LINE + u"(line: {})".format(line_count))
                line_buffer = line_buffer[-LINE_BUFFER_SIZE:]
                match_in_buffer = False
        else:
            if len(line_buffer) == LINE_BUFFER_SIZE:
                line_buffer.pop(0)
            line_buffer.append(l)
    if match_in_buffer:
        print(START_LINE)
        for i in line_buffer:
            print(i, end="")
        print(END_LINE + u"(line: {})".format(line_count))


def transformBulletsView(s):
    l = re.compile("[0-9]+\. ").split(s)
    for i in range(1, len(l)):
        l[i] = u"\n   \u2022 " + l[i]
    if len(l) > 1:
        l[0] = '\n' + l[0]
        l[-1] = l[-1] + '\n'
    return ''.join(l)



from ada import query_calc
from ada import query_memory
from ada import query_show
from ada import query_time

def process_query(args):
    if 'calc' in args:
        print(query_calc.eval_expr(*args[args.index('calc')+1:]))
    elif 'show' in args:
        query_show.process_show(*args[args.index('show')+1:])
    elif 'peek' in args or 'remember' in args or 'register' in args or 'forget' in args or 'config' in args or 'amend' in args or 'append' in args or 'pwd' in args:
        query_memory.process_memory(*args)
    elif "help" in args or "--help" in args:
        query_show.process_show("help")
    else:
        query_time.process_time_query(*args)



import os
from ada import txt_editor
# editor = txt_editor.LineFixedTxtEditor(100, ["hello"*100, "world"])
# print(editor.run())

_, EDITOR_WIDTH = os.popen('stty size', 'r').read().split()
editor = txt_editor.LineFixedTxtEditor(int(EDITOR_WIDTH) - 1, ["helllo "*3, "worrrld"], ["1:", "222222:"])
print(editor.run())

# editor = txt_editor.InteractivePrompt(100, "yes or no: ")
# print(editor.run())



import copy
import curses
import curses.textpad
import os
from threading import Thread

VIEW_STR = "[VIEW MODE]"
CMD_STR = "[CMD MODE]"
INSERT_STR = "[INSERT MODE]"
UNKNOWN_STR = "[UNKNOW STATE!!!]"

MOVE_KEYS = [curses.KEY_UP, curses.KEY_DOWN, curses.KEY_LEFT, curses.KEY_RIGHT]
BORDER_LINES = 2


class Line:
    def __init__(self, content_pair, width):
        self.content_ = content_pair[0]
        self.ineditable_content_ = content_pair[1]
        self.width_ = width


    def get_content(self):
        return copy.copy(self.content_)


    def get_ineditable_content(self):
        return copy.copy(self.ineditable_content_)


    def get_breaks(self):
        if len(self.content_) == 0 or len(self.content_) % self.width_ > 0:
            return len(self.content_) // self.width_ + 1
        else:
            return len(self.content_) // self.width_


    def get_corrected_yx(self, i):
        if self.width_ > 0:
            return [i // self.width_, i % self.width_]
        else:
            return [0, 0]


    def set_width(self, width):
        self.width_ = width


    def content_len(self):
        return len(self.content_)


    def ineditable_content_len(self):
        return len(self.ineditable_content_)


    def total_len(self):
        return self.content_len() + self.ineditable_content_len()


    def to_formatted(self):
        working_content = []
        content = self.ineditable_content_ + self.content_
        acc_size = -1
        total_size = len(content)
        while acc_size < total_size:
            acc_size = max(0, acc_size)
            stretch = min(self.width_, total_size - acc_size)
            working_content.append(content[acc_size: acc_size + stretch])
            acc_size = min(total_size, acc_size + self.width_)
        return '\n'.join(working_content)


    def to_string(self):
        return self.ineditable_content_ + self.content_


    # del a char in front of n th char
    def delete(self, n):
        i = n - len(self.ineditable_content_)
        if i > 0 and i <= self.content_len():
            self.content_ = self.content_[0:i - 1] + self.content_[i:]
            return True
        return False


    # insert a char in front of n th char
    def insert(self, n, c):
        i = n - len(self.ineditable_content_)
        if i <= self.total_len():
            self.content_ = self.content_[0:i] + c + self.content_[i:]
            return True
        return False


class LinesManager:
    def __init__(self, contents, width):
        self.width_ = width
        self.lines_ = [Line(c, width) for c in contents]
        # (logical) cursor position
        if len(self.lines_) > 0:
            self.cur_yx_ = [0, self.lines_[0].ineditable_content_len()]
        else:
            self.cur_yx_ = [0, 0]


    def del_line(self, i):
        if i < len(self.lines_):
            del self.lines_[i]


    def to_corrected_yx(self):
        acc_lines = 0
        for i in range(0, self.cur_yx_[0]):
            acc_lines = acc_lines + self.lines_[i].get_breaks()
        yx = self.lines_[self.cur_yx_[0]].get_corrected_yx(self.cur_yx_[1])
        yx[0] = yx[0] + acc_lines
        return yx


    def move_up(self):
        if self.cur_yx_[0] > 0:
            self.cur_yx_[0] = self.cur_yx_[0] - 1
            self.cur_yx_[1] = max(self.lines_[self.cur_yx_[0]].ineditable_content_len(), min(self.cur_yx_[1], self.lines_[self.cur_yx_[0]].total_len()))


    def move_down(self):
        if self.cur_yx_[0] < len(self.lines_) - 1:
            self.cur_yx_[0] = self.cur_yx_[0] + 1
            self.cur_yx_[1] = max(self.lines_[self.cur_yx_[0]].ineditable_content_len(), min(self.cur_yx_[1], self.lines_[self.cur_yx_[0]].total_len()))


    def move_left(self):
        if self.cur_yx_[1] > self.lines_[self.cur_yx_[0]].ineditable_content_len():
            self.cur_yx_[1] = self.cur_yx_[1] - 1


    def move_right(self):
        if self.cur_yx_[1] < self.lines_[self.cur_yx_[0]].total_len():
            self.cur_yx_[1] = self.cur_yx_[1] + 1


    def delete(self):
        if self.cur_yx_[0] < len(self.lines_):
            if self.lines_[self.cur_yx_[0]].delete(self.cur_yx_[1]):
                self.move_left()


    def insert(self, c):
        if self.cur_yx_[0] < len(self.lines_):
            if self.lines_[self.cur_yx_[0]].insert(self.cur_yx_[1], c):
                self.move_right()


    def to_formatted(self):
        ctt = [l.to_formatted() for l in self.lines_]
        return '\n'.join(ctt)


    def yeild_contents(self):
        return [[l.get_content() for l in self.lines_], [l.get_ineditable_content() for l in self.lines_]]


    def to_beginning(self):
        if len(self.lines_) > 0:
            self.cur_yx_ = [0, self.lines_[0].ineditable_content_len()]
        else:
            self.cur_yx_ = [0, 0]


    def to_end(self):
        if len(self.lines_) > 0:
            self.cur_yx_ = [len(self.lines_) - 1, self.lines_[len(self.lines_) - 1].ineditable_content_len() + self.lines_[len(self.lines_) - 1].content_len()]
        else:
            self.cur_yx_ = [0, 0]


    def find_forward(self, s):
        if len(s) == 0 or len(self.lines_) == 0:
            return False
        else:
            y,x = self.cur_yx_
            counter = 0
            while counter <= len(self.lines_):
                idx = self.lines_[y].get_content().find(s, x - self.lines_[y].ineditable_content_len())
                if idx != -1:
                    self.cur_yx_[0] = y
                    self.cur_yx_[1] = self.lines_[y].ineditable_content_len() + idx
                    return True
                y = (y + 1) % len(self.lines_)
                x = self.lines_[y].ineditable_content_len()
                counter = counter + 1
            return False


    def find_backward(self, s):
        if len(s) == 0 or len(self.lines_) == 0:
            return False
        else:
            y,x = self.cur_yx_
            counter = 0
            while counter <= len(self.lines_):
                idx = self.lines_[y].get_content().rfind(s, 0, x - self.lines_[y].ineditable_content_len())
                if idx != -1:
                    self.cur_yx_[0] = y
                    self.cur_yx_[1] = self.lines_[y].ineditable_content_len() + idx
                    return True
                y = (y - 1) % len(self.lines_)
                x = self.lines_[y].ineditable_content_len() + self.lines_[y].content_len()
                counter = counter + 1
            return False


# a simple vim like editor
# Each line contains two parts: 1. non-editable content and editable content;
# doesn't support line deletion
class LineFixedTxtEditor:
    def __init__(self, width, content, ineditable_content = None):
        assert width > 0
        self.width_ = width
        self.content_ = content
        if len(self.content_) == 0:
            self.content_.append('')
        if not ineditable_content or len(content) == 0:
            self.ineditable_content_ = ['' for x in self.content_]
        else:
            assert len(content) == len(ineditable_content)
            self.ineditable_content_ = ineditable_content
        self.lines_manager = LinesManager(zip(self.content_, self.ineditable_content_), self.width_)
        ## 0 is view mode, 1 is cmd mode, 2 is insert mode
        self.state_ = 0
        self.cmd_buff_ = ""
        self.hidden_cmd_buff_ = ""
        self.stay_editing_ = True
        self.dump_backup_ = False
        # cursor pos
        self.yx_ = self.lines_manager.to_corrected_yx()
        self.border_lines_ = BORDER_LINES
        self.yx_[0] = self.yx_[0] + self.border_lines_


    def _render_state(self, window):
        if self.state_ == 0:
            stt_str = VIEW_STR
        elif self.state_ == 1:
            stt_str = CMD_STR
        elif self.state_ == 2:
            stt_str = INSERT_STR
        else:
            stt_str = UNKNOWN_STR
        window.addstr(0, 0, stt_str + '\n' + self.cmd_buff_)


    def _render_content(self, window):
        yx = curses.getsyx()
        window.addstr(self.border_lines_, 0, self.lines_manager.to_formatted())


    def _render(self, window):
        self._render_state(window)
        self._render_content(window)


    def _take_and_process_view(self, window):
        c = window.getch()
        if c == ord('/'):
            self.cmd_buff_ = "/"
            self.cmd_hidden_cmd_buff_ = ""
            self.state_ = 1
            self.yx_ = [self.border_lines_ - 1, len(self.cmd_buff_)]
        elif c == ord('n') and len(self.cmd_hidden_cmd_buff_) > 0:
            self.lines_manager.move_right()
            if self.lines_manager.find_forward(self.cmd_hidden_cmd_buff_):
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
            else:
                self.lines_manager.move_left()
        elif c == ord('N') and len(self.cmd_hidden_cmd_buff_) > 0:
            if self.lines_manager.find_backward(self.cmd_hidden_cmd_buff_):
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
        elif c == ord('i'):
            self.state_ = 2
        elif c == ord(':'):
            self.state_ = 1
            self.cmd_buff_ = ":"
            self.yx_ = [self.border_lines_ - 1, len(self.cmd_buff_)]
        elif c in MOVE_KEYS:
            self._move_cursor(c)
        elif c == ord('g'):
            if self.hidden_cmd_buff_ == "g":
                self.hidden_cmd_buff_ = ""
                self.lines_manager.to_beginning()
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
            else:
                self.hidden_cmd_buff_ = "g"
        elif c == ord('G'):
            if self.hidden_cmd_buff_ == "G":
                self.hidden_cmd_buff_ = ""
                self.lines_manager.to_end()
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
            else:
                self.hidden_cmd_buff_ = "G"


    def _take_and_process_cmd(self, window):
        c = window.getch()
        if c >= 32 and c <= 126:
            self.cmd_buff_ = self.cmd_buff_ + chr(c)
            self.yx_ = [self.border_lines_ - 1, len(self.cmd_buff_)]
        elif c == 10: # ENTER
            if self.cmd_buff_ == ":wq":
                self.stay_editing_ = False
                self.cmd_buff_ = ""
            elif self.cmd_buff_ == ":w":
                window.clear()
                self._reset_state()
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
            elif self.cmd_buff_ == ":q":
                self.stay_editing_ = False
                self.cmd_buff_ = ""
                self.dump_backup_ = True
            elif self.cmd_buff_[0] == '/':
                self.cmd_hidden_cmd_buff_ = self.cmd_buff_[1:]
                self.lines_manager.find_forward(self.cmd_hidden_cmd_buff_)
                yx = self.lines_manager.to_corrected_yx()
                yx[0] = yx[0] + self.border_lines_
                self.yx_ = yx
                window.clear()
                self._reset_state()
        elif c == 27: # ESC
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
            window.clear()
            self._reset_state()
        elif c == 127:
            window.clear()
            self.cmd_buff_ = self.cmd_buff_[:-1]
            self.yx_ = [self.border_lines_ - 1, len(self.cmd_buff_)]


    def _take_and_process_insert(self, window, ch = None):
        if not ch:
            c = window.getch()
        else:
            c = ch
        if c == 27: # ESC
            window.clear()
            self._reset_state()
        elif c in MOVE_KEYS:
            self._move_cursor(c)
        elif c == 127: # delete
            # we delete char with idx lc[1] - 1 of the line
            self.lines_manager.delete()
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
            window.clear()
        elif c >= 32 and c <= 126:
            self.lines_manager.insert(chr(c))
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
            window.clear()


    def _reset_state(self):
        self.state_ = 0
        self.cmd_buff_ = ""
        self.stay_editing_ = True


    def _move_cursor(self, move):
        assert move in MOVE_KEYS, "UNKNOWN MOVE KEYS"
        if move == MOVE_KEYS[0]:
            self.lines_manager.move_up()
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
        elif move == MOVE_KEYS[1]:
            self.lines_manager.move_down()
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
        elif move == MOVE_KEYS[2]:
            self.lines_manager.move_left()
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx
        elif move == MOVE_KEYS[3]:
            self.lines_manager.move_right()
            yx = self.lines_manager.to_corrected_yx()
            yx[0] = yx[0] + self.border_lines_
            self.yx_ = yx


    def _refresh_and_wait_on_input(self, scr):
        self._render(scr)
        scr.move(*self.yx_)
        if self.state_ == 0:
            self._take_and_process_view(scr)
        elif self.state_ == 1:
            self._take_and_process_cmd(scr)
        elif self.state_ == 2:
            self._take_and_process_insert(scr)


    def _refresh_window(self, scr):
        self._render(scr)
        scr.move(*self.yx_)
        scr.refresh()


    def run(self):
        self._reset_state()
        scr = curses.initscr()
        curses.noecho()
        scr.keypad(True)
        scr.clearok(True)
        while self.stay_editing_:
            self._refresh_and_wait_on_input(scr)
        scr.clear()
        curses.endwin()
        self._reset_state()
        if not self.dump_backup_:
            return self.lines_manager.yeild_contents()
        return [self.content_, self.ineditable_content_]


class InteractivePrompt(LineFixedTxtEditor):
    def __init__(self, width, msg):
        LineFixedTxtEditor.__init__(self, width, [''], [msg])
        self.border_lines_ = 0
        self.state_ = 2
        self.yx_ = [0, len(msg)]

    def _render_state(self, window):
        pass

    def _reset_state(self):
        pass

    def _take_and_process_insert(self, window, ch = None):
        c = window.getch()
        if c == 10: # ENTER
            self.stay_editing_ = False
        else:
            LineFixedTxtEditor._take_and_process_insert(self, window, c)



import datetime, os, re
from ada.query_time import WEEKDAY, MONTH

DAY_FORMAT = '{:^5}'
NUM_FORMAT = '[0-9]+'

def process_show(*args):
    this_dir, _ = os.path.split(__file__)
    if 'yourself' in args:
        file_path = os.path.join(this_dir, 'data', 'ada.dat')
        print(open(file_path, 'r').read())
    if 'help' in args:
        file_path = os.path.join(this_dir, 'data', 'help.txt')
        print(open(file_path, 'r').read())
    elif 'calendar' in args:
        ym = getYearMonth(args)
        print_calendar(datetime.date(ym[0], ym[1], 1))
    else:
        print('Unable to show...')

def getYearMonth(args):
    today = datetime.date.today()
    year = today.year
    month = today.month
    for arg in args:
        m = re.match(NUM_FORMAT, arg)
        if m:
            num = int(m.string[m.start():m.end()])
            if num <= 12:
                month = num
            else:
                year = num
            continue
        mcount = 1
        for mon in MONTH:
            m = re.match(mon, arg.lower())
            if m:
                month = mcount
                break
            mcount = mcount + 1
    return (year, month)

def print_calendar(day_ptr):
    today = datetime.date.today()
    rows2print = [day_ptr.strftime(' ' * 14 + '%b %Y')]
    s = ''
    for d in WEEKDAY:
        s = s + DAY_FORMAT.format(d)
    rows2print.append(s)
    s = ''
    month = day_ptr.month
    wd = 0
    while day_ptr.month == month:
        if day_ptr.weekday() != wd:
            s = s + DAY_FORMAT.format('')
        else:
            if wd == 0 and len(s) > 0:
                rows2print.append(s)
                s = ''
            if day_ptr == today:
                s = s + DAY_FORMAT.format('*' + str(day_ptr.day) + '*')
            else:
                s = s + DAY_FORMAT.format(day_ptr.day)
            day_ptr = day_ptr + datetime.timedelta(days = 1)
        wd = (wd + 1) % 7
    if len(s) > 0:
        rows2print.append(s)
    for row in rows2print:
        print(row)



#!/usr/bin/python3

from distutils.core import setup

setup(
    name='ada_assistant',
    version='1.2.0',
    description='Ada: cmd helper that does lot of things',
    long_description='To be added when author is free...',
    long_description_content_type='text/x-rst',
    author='yuquanshan',
    license='MIT',
    packages = ['ada'],
    install_requires=[
        'pycrypto',
    ],
    python_requires='>=3',
    package_data={'ada': ['data/*.dat', 'data/*.txt']},
    entry_points={
        'console_scripts': [
            'ada=ada.main:main',
        ]
    },
)



import datetime, re, sys

WEEKDAY = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']
MONTH = ['jan', 'feb', 'mat', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
DATE_KEYWORDS = ['yes', 'tod', 'tom']
DATE_FORMAT = "[0-9]+(-[0-9]+)*"
NTH = ['first', 'second', 'third', 'forth', 'fifth']
NTH2 = ['1st', '2nd', '3rd', '4th', '5th']

class DontGetDateHint(Exception):
    pass


def process_time_query(*args):
    if len(args) == 0:
        print("No time specified")
    m = re.search(DATE_FORMAT, args[-1])
    args = [x.lower() for x in args]
    try:
        if m:
            if len(args) == 1:
                date = parse_date(args[-1])
            elif len(args) >= 3:
                start_date = parse_date(args[-1])
                if args[0] in NTH + NTH2:
                    if args[0] in NTH:
                        n = NTH.index(args[0])
                    else:
                        n = NTH2.index(args[0])
                    start_date = datetime.date(start_date.year, start_date.month, 1)
                    weekday = WEEKDAY.index(args[1][0:3])
                    date = nth_weekday(n, weekday, start_date)
                elif args[-2] in ['after', 'later']:
                    date = process_day_delta(int(args[0]), start_date)
                elif args[-2] in ['before', 'earlier']:
                    date = process_day_delta(-1 * int(args[0]), start_date)
                else:
                    raise DontGetDateHint
            else:
                raise DontGetDateHint
        elif args[-1] in ['after', 'later']:
            date = process_day_delta(int(args[0]))
        elif args[-1] in ['before', 'earlier']:
            date = process_day_delta(-1 * int(args[0]))
        elif len(args[-1]) > 2 and args[-1][0:3] in WEEKDAY:
            date = process_weekday(*args)
        elif len(args[-1]) > 2 and args[-1][0:3] in DATE_KEYWORDS:
            date = process_day_delta(DATE_KEYWORDS.index(args[-1][0:3]) - 1)
        else:
            raise DontGetDateHint
    except DontGetDateHint:
        print("I don't get those hints: " + ' '.join(args))
        sys.exit(1)
    except ValueError:
        print("I don't get those hints: " + ' '.join(args))
        sys.exit(1)
    print_date(date)


def parse_date(date_str):
    ymd = date_str.split('-')
    today = datetime.date.today()
    try:
        ymd = [int(x) for x in ymd]
    except ValueError:
        raise DontGetDateHint
    if len(ymd) == 1:
        # only year is specified
        return datetime.date(int(ymd[0]), 1, 1)
    elif len(ymd) == 2:
        # year or date not specified, we will see...
        try:
            # maybe year not specified?
            return datetime.date(today.year, *ymd)
        except ValueError:
            # or maybe day not specified?
            try:
                return datetime.date(*ymd, day = 1)
            except ValueError:
                raise DontGetDateHint
    elif len(ymd) == 3:
        return datetime.date(*ymd)
    else:
        raise DontGetDateHint


def nth_weekday(n, weekday, start_date):
    start_weekday = start_date.weekday()
    curr_month = start_date.month
    oldn = n
    if start_weekday > weekday:
        skip_to = 7 + weekday - start_weekday
    else:
        skip_to = weekday - start_weekday
    ret = start_date + datetime.timedelta(days = skip_to)
    while n > 0:
        ret = ret + datetime.timedelta(days = 7)
        n = n - 1
    if curr_month != ret.month:
        print(' '.join([NTH2[oldn], WEEKDAY[weekday], "doesn't exist in ", start_date.strftime('%b %Y')]))
        exit(1)
    else:
        return ret


def process_day_delta(delta, day=datetime.date.today()):
    return day + datetime.timedelta(days=delta)


def process_weekday(*args):
    nextc = args.count('next')
    lastc = args.count('last')
    today = datetime.date.today()
    if today.weekday() > WEEKDAY.index(args[-1][0:3]) and lastc > 0:
        lastc = lastc - 1
    elif today.weekday() < WEEKDAY.index(args[-1][0:3]) and nextc > 0:
        nextc = nextc - 1
    return today - datetime.timedelta(days = (lastc - nextc) * 7 + today.weekday() - WEEKDAY.index(args[-1][0:3]))


def print_date(date):
    print(date.strftime('%Y-%m-%d %A'))
